from pathlib import Path
from contextlib import contextmanager
from requests.models import Response
from requests.sessions import Session
import json as JSON
import time
import re
import logging
import inspect
import warnings
from .package_info import PackageEnvInfo
from ..tools.small_tool import get_log_info


class Request(PackageEnvInfo, Session):

    def __init__(self, test_env: str = None):
        """
        Build a Request object
        :param test_env: if you want to use specific test env in this instance
        """
        warnings.filterwarnings('ignore', message='Unverified HTTPS request')
        # self.session = requests.sessions.Session()
        super().__init__()
        self.url = None
        self._env = test_env if test_env else super().TestEnv
        self.logger = logging.getLogger(__name__)

    def re_param(self, url, **kwargs):
        """
        please add own data in env_info.py and update to cdc_automation.package_info.PackageEnvInfo
        """
        if "http" not in url:
            self.url = super().build_url(url, self._env)
        else:
            self.url = url

        if "headers" not in kwargs:
            # self.session.headers = super().build_header(url, self._env)
            self.headers = super().build_header(url, self._env)

    def _custom_logger(self, response):
        def repl(matchobj):
            if matchobj.group(0) == "\\":
                return ""
            else:
                return matchobj.group(0)[:-1] + "."
        caller_frame = inspect.currentframe().f_back.f_back
        frame_info = inspect.getframeinfo(caller_frame)
        file_path = re.search(rf"{re.escape(str(super().project_root_dir))}(.+)", frame_info.filename).group(1)
        function_name = frame_info.function
        line_number = frame_info.lineno

        matched_regex = ".*?\\\\"
        response_log = self.format_res_log(response)
        self.logger.debug(
            f'{re.sub(matched_regex, repl, file_path)}:{function_name}[{line_number}] {JSON.dumps(response_log)}',
            extra=response_log
        )

    def get(self, url, logged=True, **kwargs):
        self.re_param(url, **kwargs)
        # res = self.session.get(self.url, **kwargs, verify=False if 'https://' in self.url else None)
        res = super().get(self.url, **kwargs, verify=False if 'https://' in self.url else None)
        if logged is True:
            self._custom_logger(res)
        return res

    def post(self, url, data=None, json=None, logged=True, **kwargs):
        self.re_param(url, **kwargs)
        # res = self.session.post(self.url, data, json, **kwargs, verify=False if 'https://' in self.url else None)
        res = super().post(self.url, data, json, **kwargs, verify=False if 'https://' in self.url else None)
        if logged is True:
            self._custom_logger(res)
        return res

    def put(self, url, data=None, **kwargs):
        self.re_param(url, **kwargs)
        # res = self.session.put(self.url, data, **kwargs, verify=False if 'https://' in self.url else None)
        res = super().put(self.url, data, **kwargs, verify=False if 'https://' in self.url else None)
        self._custom_logger(res)
        return res

    def delete(self, url, **kwargs):
        self.re_param(url, **kwargs)
        # res = self.session.delete(self.url, **kwargs, verify=False if 'https://' in self.url else None)
        res = super().delete(self.url, **kwargs, verify=False if 'https://' in self.url else None)
        self._custom_logger(res)
        return res

    @staticmethod
    def format_res_log(res: Response) -> dict:
        """
        Change Response Obj to log format

        :param res: Response Obj
        :return: log format as dict
        """
        if res.request.method == "GET":
            return {
                "request": {
                    "url": str(res.request.url),
                    "headers": dict(res.request.headers)
                },
                "response": {
                    "headers": dict(res.headers),
                    "body": res.json() if res.headers.get("content-type") and "json" in res.headers.get("content-type") else None
                }
            }
        elif res.request.method == "POST":
            return {
                "request": {
                    "url": str(res.request.url),
                    "headers": dict(res.request.headers),
                    "body": JSON.loads(res.request.body) if res.request.headers.get("content-type") == "application/json" else None
                },
                "response": {
                    "headers": dict(res.headers),
                    "body": res.json() if res.headers.get("content-type") and "json" in res.headers.get("content-type") else None
                }
            }
        elif res.request.method == "PUT":
            return {
                "request": {
                    "url": str(res.request.url),
                    "headers": dict(res.request.headers),
                    "body": JSON.loads(res.request.body)
                },
                "response": {
                    "headers": dict(res.headers),
                    "body": res.json() if res.headers.get("content-type") and "json" in res.headers.get("content-type") else None
                }
            }
        elif res.request.method == "DELETE":
            return {
                "request": {
                    "url": str(res.request.url),
                    "headers": dict(res.request.headers)
                },
                "response": {
                    "headers": dict(res.headers),
                    "body": res.json() if res.headers.get("content-type") and "json" in res.headers.get("content-type") else None
                }
            }


class RequestContext(Session):

    def __init__(self, base_url: str = "", default_headers: dict[str, str] = None, verify: bool = None, token_refresher: object = None, logged: bool = True):
        """

        :param base_url:
        :param default_headers:
        :param verify:
        :param token_refresher: will pass dict key {"token", "expire_at", "refresh_token"} param to object and object must return tuple (new_token, expires_in, refresh_token)
        :param logged:
        """
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.base_url = base_url
        self.headers.update(default_headers or {})
        self.verify = verify
        self.logged = logged
        self.token_refresher = token_refresher
        self.token_info = {
            "token": self.headers.get("Authorization"),  # 初始 token
            "expire_at": None,  # 過期時間（timestamp）
            "refresh_token": None  # 交換用token
        }

    def request(self, method, url, *args, **kwargs):
        # combine URL if no http in url by this time request
        url = url if re.search(r"^http", url) else self.base_url + url

        # add default verify value in kwargs if no "verify" value by this time request
        try:
            kwargs["verify"]
        except KeyError as e:
            kwargs["verify"] = self.verify

        # 在送出前自動刷新 token
        self._refresh_token_if_needed()

        # send request
        response = super().request(method, url, *args, **kwargs)

        # record log from response
        if self.logged is True:
            self._custom_logger(response)
        return response

    def _custom_logger(self, response):
        file_path, func_name, line_no = get_log_info(depth=4)
        response_log = self.format_res_log(response)
        self.logger.debug(
            f'{file_path}:{func_name}[{line_no}] {JSON.dumps(response_log)}',
            extra=response_log
        )

    @contextmanager
    def temp_logged_status(self, temp_logged_status: bool = False):
        """
        set temporary logged status by with statement

        with request_context.temp_logged_status(False):
            print(request_context.logged)
        """
        original = self.logged
        self.logged = temp_logged_status
        try:
            yield
        finally:
            self.logged = original

    @staticmethod
    def format_res_log(res: Response) -> dict:
        """
        Change Response Obj to log format

        :param res: Response Obj
        :return: log format as dict
        """
        if res.request.method == "GET":
            return {
                "request": {
                    "url": str(res.request.url),
                    "headers": dict(res.request.headers)
                },
                "response": {
                    "headers": dict(res.headers),
                    "body": res.json() if res.headers.get("content-type") and "json" in res.headers.get("content-type") else None
                }
            }
        elif res.request.method == "POST":
            return {
                "request": {
                    "url": str(res.request.url),
                    "headers": dict(res.request.headers),
                    "body": JSON.loads(res.request.body) if res.request.headers.get("content-type") == "application/json" else None
                },
                "response": {
                    "headers": dict(res.headers),
                    "body": res.json() if res.headers.get("content-type") and "json" in res.headers.get("content-type") else None
                }
            }
        elif res.request.method == "PUT":
            return {
                "request": {
                    "url": str(res.request.url),
                    "headers": dict(res.request.headers),
                    "body": JSON.loads(res.request.body)
                },
                "response": {
                    "headers": dict(res.headers),
                    "body": res.json() if res.headers.get("content-type") and "json" in res.headers.get("content-type") else None
                }
            }
        elif res.request.method == "DELETE":
            return {
                "request": {
                    "url": str(res.request.url),
                    "headers": dict(res.request.headers)
                },
                "response": {
                    "headers": dict(res.headers),
                    "body": res.json() if res.headers.get("content-type") and "json" in res.headers.get("content-type") else None
                }
            }

    def set_token(self, token: str, expires_in: int = None, refresh_token: str = None):
        """設定 token 與過期時間"""
        self.token_info["token"] = token
        self.token_info["expire_at"] = time.time() + expires_in if expires_in else None
        self.token_info["refresh_token"] = refresh_token
        self.headers["Authorization"] = f"Bearer {token}"

    def _is_token_expired(self) -> bool:
        """檢查 token 是否過期"""
        expire_at = self.token_info.get("expire_at")
        if not expire_at:
            return False  # 沒設定過期時間 → 視為不會過期
        return time.time() + 10 >= expire_at

    def _refresh_token_if_needed(self):
        """如有設定 token_refresher，則在過期時自動刷新"""
        if self._is_token_expired() and callable(self.token_refresher):
            new_token, expires_in, refresh_token = self.token_refresher(**self.token_info)
            self.set_token(new_token, expires_in, refresh_token)



from threading import Timer
from functools import reduce
from decimal import Decimal, ROUND_HALF_UP
from pathlib import Path
import operator
import jsonschema
import re
import random
import string
import inspect


def json_schema_validate(data: dict, schema: dict, index=""):
    """
    Validate an instance data under the given schema

    :param data: response json as dict
    :param schema: api json schema as dict
    :param index: input data index for trace used, example: userId
    """
    validator = jsonschema.Draft202012Validator(schema, format_checker=jsonschema.FormatChecker())

    errors = validator.iter_errors(data)
    schema_errors = ""
    for error in errors:
        schema_errors += f"\n{error.absolute_path}, {error.message}, {index}"
    if schema_errors:
        # raise jsonschema.ValidationError(schema_errors)
        raise AssertionError(schema_errors)


def find_column(sheet, index_row, find_str):
    for cell in sheet[str(index_row)]:
        if find_str == cell.value:
            return cell


def str_gen(char=string.ascii_letters + string.digits, size=20):
    return ''.join(random.choices(char, k=size))


def custom_round(number, digits=0):
    d = Decimal(str(number))
    factor = Decimal(f'1e{-digits}')
    return float(d.quantize(factor, rounding=ROUND_HALF_UP))


class RepeatingTimer(Timer):

    def __init__(self, interval, function, args=None, kwargs=None, recursion=False):
        """
        Call a function after a specified number of seconds till cancel:
        t = Timer(30.0, f, args=None, kwargs=None) t.start() t.cancel() # stop the timer's action if it's still waiting
        :param interval:
        :param function:
        :param args:
        :param kwargs:
        :param recursion: use function returned value to next call
        """
        super().__init__(interval, function, args=args, kwargs=kwargs)
        self.recursion = recursion

    def run(self):
        self.finished.wait(self.interval)
        while not self.finished.is_set():
            if self.recursion:
                func_result = self.function(*self.args, **self.kwargs)
                if isinstance(func_result, dict):
                    self.kwargs = func_result
                elif isinstance(func_result, list | tuple):
                    self.args = func_result
                else:
                    raise ValueError("in recursion=True, function must return dict, list or tuple")
            else:
                self.function(*self.args, **self.kwargs)
            self.finished.wait(self.interval)


class DictTool:
    @staticmethod
    def transfer_json_path(path: str):
        """
        transfer json path to list
        :param path: json path, ex: "TRANRS[0].district[0].districtName"
        :return: list of json path , ex: ['TRANRS', 0, 'district', 0, 'districtName']
        """
        return [int(s) if s.isdigit() else s for s in re.split(r"\.", re.sub(r"\[(\d+)]", r".\g<1>", path))]

    @classmethod
    def get_by_path(cls, root: dict, items: list):
        """
        Access a nested object in root by item sequence.
        :param root: api request payload as dict format
        :param items: key path, ex: ['TRANRS', 0, 'district', 0, 'districtName']
        :return:
        """
        return reduce(operator.getitem, items, root)

    @classmethod
    def set_by_path(cls, root: dict, items: list, value):
        """
        Set a value in a nested object in root by item sequence.
        :param root: api request payload as dict format
        :param items: key path, ex: ['TRANRS', 0, 'district', 0, 'districtName']
        :param value: value of key that you want to set
        :return:
        """
        cls.get_by_path(root, items[:-1])[items[-1]] = value


def _find_project_root(possible_markers=None):
    if possible_markers is None:
        possible_markers = ["pyproject.toml", "setup.py", "requirements.txt", ".git"]

    current_path = Path(__file__).resolve()
    for parent in current_path.parents:
        for marker in possible_markers:
            if (parent / marker).exists():
                return parent
    # 如果都找不到，就回傳最上層資料夾
    return current_path.parents[-1]


_PROJECT_ROOT = _find_project_root()


def get_log_info(depth=1, project_root=_PROJECT_ROOT):
    """

    :param depth: counts of f_back in caller_frame.f_back
    :param project_root:
    :return: file_path, function_name, line_number
    """
    def repl(matchobj):
        if matchobj.group(0) == "\\":
            return ""
        else:
            return matchobj.group(0)[:-1] + "."

    caller_frame = inspect.currentframe()
    for i in range(depth):
        caller_frame = caller_frame.f_back
    # caller_frame = inspect.currentframe().f_back.f_back
    frame_info = inspect.getframeinfo(caller_frame)
    file_path = re.search(rf"{re.escape(str(project_root))}(.+)", frame_info.filename).group(1)
    function_name = frame_info.function
    line_number = frame_info.lineno

    matched_regex = ".*?\\\\"
    return re.sub(matched_regex, repl, file_path), function_name, line_number
